import unittest

import numpy

from pytolemaic.utils.dmd import DMD


class TestDMD(unittest.TestCase):

    def _func(self, x, is_classification):
        y = numpy.sum(x, axis=1) + x[:, 0] - x[:, 1]
        if is_classification:
            y = numpy.round(y, 0).astype(int)
        return y.reshape(-1, 1)

    def get_data(self, is_classification):
        x = numpy.random.rand(1000, 10)

        # 1st is double importance, 2nd has no importance
        y = self._func(x, is_classification=is_classification)
        return DMD(x=x, y=y,
                   columns_meta={DMD.FEATURE_NAMES: ['f_' + str(k) for k in
                                                     range(x.shape[1])]},
                   samples_meta={'sample_weight': numpy.random.rand(x.shape[0])})

    def test_properties(self):
        dmd = self.get_data(is_classification=False)

        self.assertEqual(dmd.n_features, 10)
        self.assertEqual(dmd.n_samples, 1000)
        self.assertTrue(numpy.all(
            dmd.target == self._func(dmd.values, is_classification=False)))

        self.assertListEqual(dmd.feature_names,
                             ['f_' + str(k) for k in range(dmd.n_features)])

    def test_append(self):

        dmd1 = self.get_data(is_classification=False)
        dmd2 = self.get_data(is_classification=True)

        self.assertEqual(dmd1.n_samples, dmd2.n_samples)

        dmd1.append(dmd2, axis=0)

        self.assertEqual(dmd1.n_samples, 2*dmd2.n_samples)
        self.assertEqual(dmd1._x.shape[0], dmd1._y.shape[0])
        self.assertEqual(dmd1._x.shape[0], dmd1._samples_meta.shape[0])

        self.assertEqual(dmd1.n_features, dmd2.n_features)

    def test_concat(self):
        dmd1 = self.get_data(is_classification=False)
        dmd2 = self.get_data(is_classification=True)

        self.assertEqual(dmd1.n_samples, dmd2.n_samples)

        dmd = DMD.concat([dmd1, dmd2], axis=0)

        self.assertEqual(dmd.n_samples, 2 * dmd2.n_samples)
        self.assertEqual(dmd._x.shape[0], 2* dmd1._y.shape[0])
        self.assertEqual(dmd._x.shape[0], 2* dmd1._samples_meta.shape[0])

        self.assertEqual(dmd.n_features, dmd2.n_features)



import unittest

from examples.toy_examples import example_quality_report, example_scoring_report, example_sensitivity_analysis, \
    example_prediction_uncertainty, example_prediction_explanation

from examples import print_help_manual

class TestToyExamples(unittest.TestCase):

    def test_example_scoring_report(self):
        example_scoring_report.run()

    def test_example_sensitivity_analysis(self):
        example_sensitivity_analysis.run()

    def test_example_prediction_uncertainty(self):
        example_prediction_uncertainty.run()

    def test_example_quality_report(self):
        example_quality_report.run()

    def test_example_prediction_explanation(self):
        example_prediction_explanation.run()

    def test_example_help(self):
        print_help_manual.run()


import logging

import numpy
import numpy as np

from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity_reports import SensitivityOfFeaturesReport, \
    SensitivityStatsReport, SensitivityVulnerabilityReport, SensitivityFullReport, SensitivityTypes
from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.general import GeneralUtils
from pytolemaic.utils.metrics import Metrics


class SensitivityAnalysis():
    def __init__(self):
        self.metrics = Metrics.supported_metrics()
        self.model_support_dmd = None
        self.max_samples_to_use = 20000
        self.low_sensitivity_threshold = 0.05
        self.very_low_sensitivity_threshold = 1e-4

    @classmethod
    def shuffle_x(cls, x, index, dmd_train=None, seed=0):
        rs = np.random.RandomState(seed)
        if dmd_train is None:
            x[:, index] = rs.choice(x[:, index], len(x), replace=True)
        else:
            x[:, index] = rs.choice(dmd_train.values[:, index], len(x), replace=True)
        return x

    @classmethod
    def get_shuffled_x(cls, dmdx: DMD, index=None, dmd_train=None, method=SensitivityTypes.shuffled, seed=0,
                       model_support_dmd=False):
        if index is None:
            return dmdx.values

        x_copy = numpy.copy(dmdx.values)
        if method == SensitivityTypes.shuffled:
            x_copy = cls.shuffle_x(x_copy, dmd_train=dmd_train, index=index, seed=index + seed)
        if method == SensitivityTypes.missing:
            x_copy[:, index] = np.nan

        if model_support_dmd:
            return DMD(x=x_copy,
                       samples_meta=dmdx._samples_meta,
                       columns_meta=dmdx._columns_meta,
                       splitter=dmdx.splitter)
        else:
            return x_copy

    def sensitivity_analysis(self, model, dmd_test: DMD, metric,
                             dmd_train=None,
                             method=SensitivityTypes.shuffled, raw_scores=False,
                             y_pred=None) -> SensitivityOfFeaturesReport:

        self.model_support_dmd = GeneralUtils.dmd_supported(model, dmd_test)
        x = dmd_test if self.model_support_dmd else dmd_test.values

        y_pred = y_pred or model.predict(x)
        ytest = dmd_test.target

        score_function = self.metrics[metric].function
        if metric in ['auc', 'logloss'] and ytest is not None:
            base_score = score_function(ytest, model.predict_proba(x))
            y_pred = ytest
        else:
            base_score = 0
            y_pred = y_pred

        predict_function = model.predict_proba if self.metrics[
            metric].is_proba \
            else model.predict

        scores = {}
        for i, name in enumerate(dmd_test.feature_names):
            if dmd_test.n_samples > self.max_samples_to_use:
                rs = numpy.random.RandomState(i)
                subset = rs.permutation(dmd_test.n_samples)[:self.max_samples_to_use]
                dmd_test_ = dmd_test.split_by_indices(subset)
                y_pred_ = y_pred[subset]
            else:
                dmd_test_ = dmd_test
                y_pred_ = y_pred

            shuffled_x = self.get_shuffled_x(dmd_test_, i, dmd_train=dmd_train, method=method,
                                             model_support_dmd=self.model_support_dmd)
            shuffled_pred = predict_function(shuffled_x)

            if base_score > 0:
                scores[name] = 1 - abs(base_score - score_function(y_pred_,
                                                                   shuffled_pred))  # higher difference - more impact so add 1- in front
            else:
                scores[name] = score_function(y_pred_, shuffled_pred)  # higher score - less impact

        if raw_scores:
            # description = "The raw scores of how each feature affects the model's predictions."
            return SensitivityOfFeaturesReport(method=method, sensitivities=scores,
                                               stats_report=self._sensitivity_stats_report(scores))

        # higher score / lower loss means the shuffled feature did less impact
        if self.metrics[metric].is_loss:
            impact = scores
        else:
            impact = {name: 1 - score for name, score in scores.items()}

        total_impact = sum([score for score in impact.values()])
        impact = {name: float(score / total_impact) for
                  name, score in impact.items()}
        impact = GeneralUtils.round_values(impact)

        # description="The impact of each feature on model's predictions. "
        #             "Higher value mean larger impact (0 means no impact at all). "
        #             "Values are normalized to 1.")
        return SensitivityOfFeaturesReport(method=method, sensitivities=impact,
                                           stats_report=self._sensitivity_stats_report(sensitivities=impact))

    def _sensitivity_stats_report(self, sensitivities: dict) -> [SensitivityStatsReport, None]:
        if not sensitivities:
            return None

        sensitivity = np.abs(np.array(list(sensitivities.values())))
        n_features = len(sensitivity)

        low_sensitivity = max(sensitivity) * self.low_sensitivity_threshold
        very_low_sensitivity = max(sensitivity) * self.very_low_sensitivity_threshold
        zero_sensitivity = 0

        n_low = int(np.sum(sensitivity < low_sensitivity))
        n_very_low = int(np.sum(sensitivity < very_low_sensitivity))
        n_zero = int(np.sum(sensitivity <= zero_sensitivity))

        return SensitivityStatsReport(n_features=n_features,
                                      n_low=n_low,
                                      n_very_low=n_very_low,
                                      n_zero=n_zero)

    def _vulnerability_report(self, shuffled_sensitivity: SensitivityOfFeaturesReport,
                              missing_sensitivity: SensitivityOfFeaturesReport) -> SensitivityVulnerabilityReport:
        # lower is better
        stats = shuffled_sensitivity.stats_report

        leakage = self._leakage(n_features=stats.n_features,
                                n_very_low=stats.n_very_low)
        too_many_features = self._too_many_features(n_features=stats.n_features,
                                                    n_low=stats.n_low,
                                                    n_very_low=stats.n_very_low,
                                                    n_zero=stats.n_zero)
        imputation = self._imputation_score(
            shuffled=shuffled_sensitivity,
            missing=missing_sensitivity)

        return SensitivityVulnerabilityReport(imputation=imputation,
                                              too_many_features=too_many_features,
                                              leakage=leakage)

    def _leakage(self, n_features, n_very_low, **kwargs):
        """
        measure the chance for data leakage - strong data leakage cause only few features to contribute to the model.
        :param shuffled_sensitivity:
        :return:
        """

        if n_features < 2:
            return 0

        n_non_zero = n_features - n_very_low

        return np.power(n_very_low / n_features, n_non_zero - 1)

    def _too_many_features(self, n_features, n_low, n_very_low, n_zero, **kwargs):
        """
        Many features with low sensitivity indicate the model relies on non-informative feature. This may cause overfit.
        higher value when there are many features with low contribution
        :param shuffled_sensitivity:
        :return:
        """

        return max((n_low-n_zero) / n_features, np.sqrt((n_very_low-n_zero) / n_features))

    def _imputation_score(self, shuffled: SensitivityOfFeaturesReport, missing: SensitivityOfFeaturesReport):
        """
        missing sensitivity should (more or less) match shuffled in impact.
        If it does not match, it may mean there is an issue with imputation
        :param shuffled:
        :param missing:
        :return:
        """
        if not missing:
            return 0

        deltas = numpy.abs([shuffled.sensitivities[i] - missing.sensitivities[i] for i in shuffled.sensitivities])
        deltas = deltas[deltas >= max(deltas) * 1e-3]
        if max(abs(deltas)) == 0:
            return 0

        score = np.mean(deltas) / max(deltas)
        return score

    def calculate_sensitivity(self, model, dmd_test: DMD, metric: str, dmd_train=None, **kwargs):
        dmd_test = dmd_test or dmd_train
        self.shuffled_sensitivity = self.sensitivity_analysis(
            model=model,
            dmd_train=dmd_train,
            dmd_test=dmd_test,
            metric=metric,
            method=SensitivityTypes.shuffled,
            raw_scores=False)

        try:
            self.missing_sensitivity = self.sensitivity_analysis(
                model=model,
                dmd_test=dmd_test,
                dmd_train=dmd_train,
                metric=metric,
                method=SensitivityTypes.missing,
                raw_scores=False)
        except:
            logging.error(
                "Failed to calculate sensitivity with {} method. This is expected if your model cannot handle missing values.".format(
                    SensitivityTypes.missing))

            self.missing_sensitivity = None

    def sensitivity_report(self, **kwargs) -> SensitivityFullReport:

        vulnerability_report = self._vulnerability_report(
            shuffled_sensitivity=self.shuffled_sensitivity,
            missing_sensitivity=self.missing_sensitivity)

        return SensitivityFullReport(
            shuffle_report=self.shuffled_sensitivity,
            missing_report=self.missing_sensitivity,
            vulnerability_report=vulnerability_report
        )



import unittest
from pprint import pprint

import numpy

from pytolemaic.analysis_logic.dataset_analysis.dataset_analysis import DatasetAnalysis
from pytolemaic.utils.constants import REGRESSION, FeatureTypes, CLASSIFICATION
from pytolemaic.utils.dmd import DMD


class TestDatasetAnalysis(unittest.TestCase):

  def setUp(self) -> None:
    rs = numpy.random.RandomState(0)
    x = rs.randn(1000, 10)

    x[:, 0:5] = numpy.round(x[:, 0:5], 0)
    x[0, 0] = 10
    x[0, 9] = 10
    x[1, 9] = -4

    y = numpy.copy(x[:, 0])

    x[10, :] = numpy.nan
    x[:, 1] = numpy.nan
    x[:700, 2] = numpy.nan

    self.dataset = DMD(x=x, y=y, columns_meta={
      DMD.FEATURE_TYPES: 5 * [FeatureTypes.categorical] + 5 * [FeatureTypes.numerical]})

  def test_count_unique_classes(self):
    da = DatasetAnalysis(CLASSIFICATION, class_count_threshold=10, outliers_n_sigma=(3, 5),
                         nan_threshold_per_col=(0.1, 0.5, 0.9), nan_threshold_per_sample=(0.1, 0.5, 0.9))
    out = da.count_unique_classes(self.dataset)
    print("count_unique_classes:")
    pprint(out)

    self.assertTrue(0 in out)
    self.assertTrue(10.0 in out[0])
    self.assertTrue(out[0][10.0] == 1)

    self.assertTrue('target' in out)
    self.assertTrue(10.0 in out['target'])
    self.assertTrue(out['target'][10.0] == 1)

    for feature in out:
      for value in out[feature].values():
        self.assertLessEqual(value, da._class_count_threshold)

  def test_count_outliers(self):
    da = DatasetAnalysis(REGRESSION, class_count_threshold=10, outliers_n_sigma=(3, 5),
                         nan_threshold_per_col=(0.1, 0.5, 0.9), nan_threshold_per_sample=(0.1, 0.5, 0.9))
    out = da.count_outliers(self.dataset)
    print("count_outliers:")
    pprint(out)

    self.assertTrue(9 in out)
    self.assertTrue('3-sigma' in out[9])
    self.assertTrue('5-sigma' in out[9])
    self.assertTrue(out[9]['5-sigma']['n_outliers'] == 1)
    self.assertTrue(out[9]['3-sigma']['expected_outliers'] == 2)
    self.assertTrue(out[9]['5-sigma']['expected_outliers'] == 0)

    self.assertTrue('target' in out)
    self.assertTrue('3-sigma' not in out['target'])
    self.assertTrue('5-sigma' in out['target'])
    self.assertTrue(out['target']['5-sigma']['n_outliers'] == 1)

  def test_count_missing_values(self):
    da = DatasetAnalysis(REGRESSION, class_count_threshold=10, outliers_n_sigma=(3, 5),
                         nan_threshold_per_col=(0.1, 0.5, 0.9), nan_threshold_per_sample=(0.1, 0.9))
    nan_cols, nan_rows = da.count_missing_values(self.dataset)
    print("count_missing_values cols:")
    pprint(nan_cols)

    print("count_missing_values rows:")
    pprint(nan_rows)

    for th in da._nan_threshold_per_feature:
      self.assertTrue(th in nan_cols)

    for th in da._nan_threshold_per_sample:
      self.assertTrue(th in nan_rows)

    self.assertEqual(nan_cols, {0.1: {1: 1.0, 2: 0.7}, 0.5: {1: 1.0, 2: 0.7}, 0.9: {1: 1.0}})
    self.assertEqual(nan_rows[0.9], {10: 1.0})
    self.assertEqual(len(nan_rows[0.1]), self.dataset.n_samples)
    for th, vdict in nan_rows.items():
      print(th, vdict)
      self.assertTrue(th <= max(vdict.values()))

  def test_dataset_analysis_report(self):
    da = DatasetAnalysis(REGRESSION, class_count_threshold=10, outliers_n_sigma=(3, 5),
                         nan_threshold_per_col=(0.1, 0.5, 0.9), nan_threshold_per_sample=(0.1, 0.9))
    report = da.dataset_analysis_report(train=self.dataset)
    pprint(report.to_dict(printable=True))
    pprint(report.to_dict_meaning())
    pprint(report.plot())
    #
    # from matplotlib import pyplot as plt
    # plt.show()



import unittest

from pytolemaic.analysis_logic.dataset_analysis.covariance_shift_report import CovarianceShiftReport
from pytolemaic.analysis_logic.dataset_analysis.dataset_analysis_report import DatasetAnalysisReport, \
    MissingValuesReport
from pytolemaic.analysis_logic.model_analysis.scoring.scoring_report import ScoringFullReport, ConfusionMatrixReport, \
    ScatterReport, ScoringMetricReport
from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity_reports import SensitivityVulnerabilityReport
from pytolemaic.analysis_logic.quality_report import TestSetQualityReport, TrainSetQualityReport, QualityReport, \
    ModelQualityReport
from pytolemaic.utils.metrics import Metrics


class TestSensitivityReport(unittest.TestCase):

    def equal_to_dict_keys(self, rep):
        d1 = rep.to_dict()
        d2 = rep.to_dict_meaning()
        return d1.keys() == d2.keys()

    def get_scoring_report(self):
        rep1 = ConfusionMatrixReport(y_true=[1, 2, 3], y_pred=[1, 2, 3])
        rep2 = ScatterReport(y_true=[1, 2, 3], y_pred=[1, 2, 3])
        rep3 = ScoringMetricReport(metric_name=Metrics.normalized_rmse.name, value=0.5, ci_low=0.25, ci_high=0.75)
        rep4 = ScoringMetricReport(metric_name=Metrics.mae.name, value=0.5, ci_low=0.25, ci_high=0.75)
        return ScoringFullReport(target_metric=Metrics.mae.name, metric_reports=[rep3, rep4],
                                 confusion_matrix=rep1, scatter=rep2)

    def get_vulnerability_report(self):
        return SensitivityVulnerabilityReport(imputation=0.5, leakage=0.5, too_many_features=0.5)

    def get_DatasetAnalysis_report(self):
        return DatasetAnalysisReport(class_counts={}, outliers_count={},
                                     missing_values_report=MissingValuesReport(nan_counts_features=0, nan_counts_samples=0, n_samples_to_show=10),
                                     covariance_shift_report=CovarianceShiftReport(covariance_shift=0.75))


    def test_to_dict_meaning(self):
        rep1 = TestSetQualityReport(scoring_report=self.get_scoring_report(), dataset_analysis_report=self.get_DatasetAnalysis_report())
        self.assertTrue(self.equal_to_dict_keys(rep1))

        rep2 = TrainSetQualityReport(self.get_vulnerability_report())
        self.assertTrue(self.equal_to_dict_keys(rep2))

        rep3 = ModelQualityReport(self.get_vulnerability_report(), self.get_scoring_report())
        self.assertTrue(self.equal_to_dict_keys(rep3))

        rep4 = QualityReport(test_quality_report=rep1, train_quality_report=rep2, model_quality_report=rep3)
        self.assertTrue(self.equal_to_dict_keys(rep3))



import logging

import numpy
try:
    from lime.lime_tabular import LimeTabularExplainer
except Exception as e:
    logging.exception("Failed to import lime module:")
    logging.error("Please try to run 'pip install lime'")
    raise

from matplotlib import pyplot as plt
from sklearn.linear_model import ElasticNet
from sklearn.utils.multiclass import unique_labels

from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.general import GeneralUtils


class ElasticNetWrapper(ElasticNet):

    def fit(self, X, y, check_input=False, sample_weight=None):
        return super(ElasticNetWrapper, self).fit(X, y, check_input)


class LimeExplainer():
    def __init__(self, kernel_width=3, n_features_to_plot=None, tol=1e-2, max_samples=256000, fillna=0):
        """

        :param kernel_width: Lime parameter
        :param n_features_to_plot: # of features to show/plot
        :param tol: desired convergence tol on explanation
        :param max_samples: limit on # of samples used to create explanation
        """

        self.kernel_width = kernel_width
        self.n_features_to_plot = n_features_to_plot
        self.predict_function = None
        self.tol = tol
        self.max_samples = max_samples
        self.fillna = fillna

    def fit(self, dmd_train: DMD, model):
        is_classification = GeneralUtils.is_classification(model)

        x = dmd_train.values.astype(float, copy=True)
        nan_mask = ~numpy.isfinite(x)
        if numpy.any(nan_mask):
            logging.warning(
                "Lime cannot handle missing values. Fillna={} was used to coerce the issue.".format(self.fillna))
            x[nan_mask] = self.fillna

        self.explainer = LimeTabularExplainer(
            training_data=x,
            mode="classification" if is_classification else "regression",
            training_labels=None,  # ???
            feature_names=dmd_train.feature_names,
            categorical_features=dmd_train.categorical_features if dmd_train.categorical_features is not None else [],
            ###
            categorical_names=dmd_train.categorical_encoding_by_icols,  ###
            kernel_width=self.kernel_width,
            kernel=None,
            # default is np.sqrt(np.exp(-(d ** 2) / kernel_width ** 2))
            verbose=False,
            class_names=dmd_train.labels,
            feature_selection='auto',
            # ??? options are 'forward_selection', 'lasso_path', 'none' or 'auto'.
            discretize_continuous=True,
            discretizer='decile',
            # -- Lime discretizers do not support nans (options are 'quartile', 'decile', 'entropy')
            sample_around_instance=True,  # default is False
            random_state=0,
            training_data_stats=None)

        self.model = model
        self.predict_function = self.model.predict_proba if is_classification else self.model.predict

        if is_classification:
            labels = dmd_train.labels
            if labels is None:
                labels = set(unique_labels(dmd_train.target))
            self.labels = numpy.arange(len(labels))
        else:
            self.labels = None

        self.n_features = dmd_train.n_features
        self.n_features_to_plot = self.n_features_to_plot or dmd_train.n_features
        self.n_features_to_plot = min(self.n_features_to_plot, self.n_features)

    def explain(self, sample: numpy.ndarray) -> [dict, None]:
        try:
            exp = self._converged_lime_explaination(sample)

            label = self.model.predict(sample.reshape(1, -1))
            return dict(exp.as_list(label=int(label)))
        except:
            logging.exception("Failed to produce lime explanation for sample {}".format(sample))
            return None

    def plot(self, sample: numpy.ndarray):
        try:
            exp = self._converged_lime_explaination(sample)

            label = self.model.predict(sample.reshape(1, -1))

            if GeneralUtils.is_classification(self.model):
                exp.as_pyplot_figure(label=int(label))
            else:
                exp.as_pyplot_figure(label=None)
                plt.title("Local explanation for predicted value of %.3g" % label)

            plt.tight_layout()
            plt.draw()
        except ValueError as e:
            logging.exception("Failed to plot Lime for instance\n{}".format(sample))

    def _lime_explaination(self, sample, num_samples=16000):
        model_regressor = ElasticNetWrapper(random_state=0, l1_ratio=0.9, alpha=1e-3, warm_start=True, copy_X=False,
                                            selection='random', tol=1e-4)

        exp = self.explainer.explain_instance(sample.ravel(),
                                              self.predict_function,
                                              labels=self.labels,
                                              num_features=self.n_features,
                                              num_samples=num_samples,
                                              model_regressor=model_regressor)

        return exp

    def _convergence_acheived(self, lower_exp, higher_exp):
        features_to_show = sorted(higher_exp.keys(), key=lambda key: abs(higher_exp[key]), reverse=True)[:self.n_features_to_plot]
        # inefficient diff, but self.n_features_to_plot is expected to be <20
        diff = {k: abs(lower_exp[k] - higher_exp[k]) for k in lower_exp if k in features_to_show}

        max_value = numpy.max(numpy.abs(list(higher_exp.values())))
        delta = numpy.array(list(diff.values())) / max_value
        if max(delta) < self.tol:
            converged = True
        else:
            converged = False
        return converged

    def _converged_lime_explaination(self, sample):
        def as_dict(exp):
            return {k: numpy.round(v, 5) for k, v in exp.as_list()}

        sample = numpy.array(sample, dtype=float, copy=True)
        nan_mask = ~numpy.isfinite(sample)
        if numpy.any(nan_mask):
            logging.warning("Lime cannot handle missing values. Fillna(0) was used to coerce the issue.")
            sample = numpy.copy(sample)
            sample[nan_mask] = self.fillna

        try:

            num_samples = min(16000, self.max_samples // 2)

            exp = self._lime_explaination(sample=sample, num_samples=num_samples)
            higher_exp = as_dict(exp)

            converged = False
            while not converged and num_samples < self.max_samples:
                num_samples *= 2
                lower_exp = higher_exp
                exp = self._lime_explaination(sample=sample, num_samples=num_samples)
                higher_exp = as_dict(exp)

                converged = self._convergence_acheived(lower_exp=lower_exp, higher_exp=higher_exp)

            if not converged:
                logging.warning("Lime explainer did not converge with {} samples".format(num_samples))

            return exp

        except ValueError as e:
            logging.exception("Failed to explain Lime for instance\n{}".format(sample))
            raise






import unittest

from pytolemaic import PyTrust, help


class TestPyTrust(unittest.TestCase):

    def test_example_usage(self):
        example = PyTrust.print_usage_example()
        for component in ['plot()', 'insights', 'to_dict()', 'to_dict_meaning()']:
            try:
                assert 'pytrust.report.{}'.format(component) in example
            except:
                print("component=", component)
                print("example=\n", example)
                raise

        # todo use eval?

    def test_init_example_usage(self):
        example = PyTrust.print_initialization_example()
        # todo use eval?

    def test_help(self):
        supported_keys = help()
        for key in supported_keys:
            print('\n\n\n*** TEST KEY:', key, '***')
            help(key=key)



import copy
import os

import numpy
import pandas
from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import LabelEncoder

from pytolemaic.utils.constants import FeatureTypes
from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.general import GeneralUtils

this_file = os.path.dirname(__file__)
adult_train_path = os.path.join(this_file, 'uci_adult_files', 'adult.data')
adult_test_path = os.path.join(this_file, 'uci_adult_files', 'adult.test')


class UCIAdult():
    def __init__(self):
        self._xtrain, self._ytrain = None, None
        self._xtest, self._ytest = None, None
        self.model = GeneralUtils.simple_imputation_pipeline(
            RandomForestClassifier(random_state=0, n_estimators=100, n_jobs=3))

    def column_names(self):
        return ['age', 'workclass', 'fnlwgt', 'education', 'education-num', 'marital-status', 'occupation',
                'relationship', 'race', 'sex', 'capital-gain', 'capital-loss', 'hours-per-week', 'native-country']

    def feature_types(self):
        num, cat = FeatureTypes.numerical, FeatureTypes.categorical
        types = [num, cat, num, cat, num, cat, cat, cat, cat, cat, num, num, num, cat]
        assert len(types) == len(self.column_names())
        return types

    def _encode_data(self, df, fit):
        df = copy.deepcopy(df)
        feature_types = self.feature_types()

        nan_mask = df.replace('?', numpy.nan).isnull()
        if fit:
            self.xencoders = []
            self.categorical_encoding = {}
            for i, col in enumerate(self.column_names()):
                if feature_types[i] == FeatureTypes.categorical:
                    le = LabelEncoder().fit(df[col])
                    self.xencoders.append(le)
                    self.categorical_encoding[col] = {i: cls_ for i, cls_ in enumerate(le.classes_)}
                else:
                    self.xencoders.append(None)

            self.yencoder = LabelEncoder().fit(df['target'])
            self._labels = {i: cls_ for i, cls_ in enumerate(self.yencoder.classes_)}
            self.categorical_encoding['target'] = {i: cls_ for i, cls_ in enumerate(self.labels)}

        for i, col in enumerate(self.column_names()):
            if feature_types[i] == FeatureTypes.categorical:
                df[col] = self.xencoders[i].transform(df[col])

        df['target'] = self.yencoder.transform(df['target'])

        df[nan_mask] = numpy.nan

        return df

    @property
    def labels(self):
        return self._labels

    @property
    def training_data(self):
        if self._xtrain is None:
            df = pandas.read_csv(adult_train_path)
            df = self._encode_data(df, fit=True)
            ytrain = df[['target']]
            xtrain = df.drop(['target'], axis=1)
            self._xtrain, self._ytrain = xtrain, ytrain

        return self._xtrain, self._ytrain

    @property
    def testing_data(self):
        if self._xtest is None:
            df = pandas.read_csv(adult_test_path)
            df = self._encode_data(df, fit=False)
            ytest = df[['target']]
            xtest = df.drop(['target'], axis=1)
            self._xtest, self._ytest = xtest, ytest

        return self._xtest, self._ytest

    def get_model(self):
        x, y = self.training_data
        self.model.fit(x, y)
        return self.model

    def as_dmd(self):
        train = DMD(x=self.training_data[0], y=self.training_data[1],
                    samples_meta=None, columns_meta={DMD.FEATURE_NAMES: self.column_names(),
                                                     DMD.FEATURE_TYPES: self.feature_types()},
                    target_labels=self.labels,
                    categorical_encoding=self.categorical_encoding)

        test = DMD(x=self.testing_data[0], y=self.testing_data[1],
                   samples_meta=None, columns_meta={DMD.FEATURE_NAMES: self.column_names(),
                                                    DMD.FEATURE_TYPES: self.feature_types()},
                   target_labels=self.labels,
                   categorical_encoding=self.categorical_encoding)
        return train, test


if __name__ == '__main__':
    adult = UCIAdult()
    df1, df2 = adult.training_data
    print(df1.describe())
    print(df2.describe())

    df1, df2 = adult.testing_data
    print(df1.describe())
    print(df2.describe())



from typing import List, Tuple

import numpy

from pytolemaic import DMD

class TransformerBase():
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.name = kwargs.get('name', type(self).__name__)

    def fit_transform(self, dmd:DMD, **kwargs)->DMD:
        raise NotImplementedError("fit_transform")

    def transform(self, dmd:DMD, **kwargs)->DMD:
        raise NotImplementedError("transform")

class EstimatorBase():
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.name = kwargs.get('name', type(self).__name__)

    def fit(self, dmd:DMD, **kwargs)->None:
        raise NotImplementedError("fit")

    def predict(self, dmd:DMD, **kwargs)->DMD:
        raise NotImplementedError("predict")

    def predict_proba(self, dmd:DMD, **kwargs)->DMD:
        raise NotImplementedError("predict_proba")


class DMDPipeline():
    def __init__(self, transformers: List[type(TransformerBase)],
                 estimator: type(EstimatorBase),
                 hyperparameters: List[dict],
                 name=None,
                 **kwargs):

        self.transformers = [transformer_class(**hps) for transformer_class, hps in zip(transformers, hyperparameters[:-1])]
        self.estimator = estimator(**hyperparameters[-1])

        self.hyperparameters = [(transformer.name, hps) for transformer, hps in zip(self.transformers, hyperparameters[:-1])]
        self.hyperparameters.append((self.estimator.name, hyperparameters[-1]))

        default_name = '>'.join([transformer.name for transformer in self.transformers]) + '>' + self.estimator.name
        self.name = name if name is not None else default_name

    def fit(self, dmd, **kwargs):
        for transformer in self.transformers:
            dmd = transformer.fit_transform(dmd, **kwargs)

        self.estimator.fit(dmd, **kwargs)

    def transform(self, dmd:DMD, **kwargs)->DMD:
        for transformer in self.transformers:
            dmd = transformer.transform(dmd, **kwargs)

        return dmd

    def predict(self, dmd:DMD, **kwargs)->DMD:
        dmd = self.transform(dmd, **kwargs)
        return self.estimator.predict(dmd, **kwargs)

    def predict_proba(self, dmd:DMD, **kwargs)->DMD:
        dmd = self.transform(dmd, **kwargs)
        return self.estimator.predict_proba(dmd, **kwargs)

if __name__ == '__main__':
    class TransformerDummy():
        def __init__(self, **kwargs):
            self.kwargs = kwargs
            self.name = kwargs.get('name', type(self).__name__)
            print('TransformerDummy', '__init__', 'name', self.name)


        def fit_transform(self, dmd: DMD, **kwargs):
            print('TransformerDummy','fit_transform','name',self.kwargs.get('name'))
            return dmd

        def transform(self, dmd: DMD, **kwargs):
            print('TransformerDummy', 'transform', 'name', self.kwargs.get('name'))
            return dmd


    class EstimatorDummy():
        def __init__(self, **kwargs):
            self.kwargs = kwargs
            self.name = kwargs.get('name', type(self).__name__)
            print('EstimatorDummy', '__init__', 'name', self.name)


        def fit(self, dmd: DMD, **kwargs):
            print('EstimatorDummy','fit','n_estimators',self.kwargs.get('n_estimators'))
            pass

        def predict(self, dmd: DMD, **kwargs):
            print('EstimatorDummy','predict','n_estimators',self.kwargs.get('n_estimators'))
            return DMD(x=dmd.values[:,0:1])

        def predict_proba(self, dmd: DMD, **kwargs):
            print('EstimatorDummy','predict_proba','n_estimators',self.kwargs.get('n_estimators'))
            return DMD(x=dmd.values[:,0:2])

    dmd_train = DMD(x=numpy.zeros((3,3)), y=numpy.zeros((3,1)))
    dmd_test = DMD(x=numpy.zeros((3,3)), y=numpy.zeros((3,1)))

    print('init model')
    model = DMDPipeline(transformers=[TransformerDummy],
                        estimator=EstimatorDummy,
                        hyperparameters=[{'name': 'Dummy'}, {'n_estimators':10}])

    print('fit model')
    model.fit(dmd_train)

    print('model predict')
    predictions = model.predict(dmd_test)

    """
    from sklearn.impute import SimpleImputer
    from sklearn.ensemble import RandomForestClassifier

    model = DMDPipeline(transformers=[SimpleImputer],
                        estimator=RandomForestClassifier,
                        hyperparameters=[{}, {'n_estimators':10}])
              
    dmd_train = DMD()
    dmd_test = DMD()
    model.fit(dmd_train)
    predictions = model.predict(dmd_test)
    score = evaluate(predictions, dmd_test)
    """


import time
from pprint import pprint

import numpy
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier

from pytolemaic import Metrics, DMD
from pytolemaic import PyTrust
from pytolemaic.utils.general import GeneralUtils
from resources.datasets.uci_adult import UCIAdult


def run(use_active_learning=True, max_samples = 12500, batch_size = 100, train_base_ratio=0.001):

    dataset = UCIAdult()
    train, test = dataset.as_dmd()

    metric = Metrics.recall.name

    unlabeled, train_base = train.split(ratio=train_base_ratio)

    print("# samples: Test: {}, train_base: {}, unlabeled:{}"
          .format(test.n_samples, train_base.n_samples, unlabeled.n_samples))


    x_samples=[]
    y_score =[]

    while train_base.n_samples < train.n_samples and unlabeled.n_samples > 100 and train_base.n_samples<max_samples:
        model = GeneralUtils.simple_imputation_pipeline(estimator=RandomForestClassifier(random_state=0))
        # model = GeneralUtils.simple_imputation_pipeline(estimator=KNeighborsClassifier(n_neighbors=5))
        # model = GeneralUtils.simple_imputation_pipeline(estimator=LogisticRegression())

        model.fit(train_base.values, train_base.target.ravel())

        pytrust = PyTrust(
            model=model,
            xtrain=train_base,
            xtest=test,
            metric=metric)

        score = pytrust.scoring_report.metric_scores['recall'].to_dict()['value']
        print("With {} samples, score is {:.3g}".format(train_base.n_samples,
                                                        score))
        x_samples.append(train_base.n_samples)
        y_score.append(score)

        uncertainty = pytrust.create_uncertainty_model('probability', do_analysis=False)
        y = uncertainty.uncertainty(unlabeled)

        if use_active_learning:
            inds = numpy.arange(len(y))
            sorted_y, sorted_inds = list(zip(*sorted(list(zip(y, inds)), key=lambda p: p[0], reverse=True)))
        else: # random sampling
            sorted_inds = numpy.random.permutation(len(y))

        batch_size = min(batch_size, len(sorted_inds)-100)
        train_base.append(unlabeled.split_by_indices(sorted_inds[:batch_size]))
        unlabeled = unlabeled.split_by_indices(sorted_inds[batch_size:])

    return x_samples, y_score

if __name__ == '__main__':
    from matplotlib import pyplot as plt

    x_samples, y_score = run(use_active_learning=True)
    plt.plot(x_samples, y_score, '.-b', label='active learning')
    x_samples, y_score = run(use_active_learning=False)
    plt.plot(x_samples, y_score, '.-r', label='random sampling')
    plt.legend()
    plt.show()



import unittest

import numpy

from pytolemaic import DMD
from pytolemaic.utils.pipeline_base import DMDPipeline, EstimatorBase, TransformerBase


class TransformerDummy(TransformerBase):
    def __init__(self, **kwargs):
        super(TransformerDummy, self).__init__(**kwargs)
        print('TransformerDummy', '__init__', 'name', self.name)

    def fit_transform(self, dmd: DMD, **kwargs):
        print('TransformerDummy', 'fit_transform', 'name', self.kwargs.get('name'))
        return dmd

    def transform(self, dmd: DMD, **kwargs):
        print('TransformerDummy', 'transform', 'name', self.kwargs.get('name'))
        return dmd


class EstimatorDummy(EstimatorBase):
    def __init__(self, **kwargs):
        super(EstimatorDummy, self).__init__(**kwargs)
        print('EstimatorDummy', '__init__', 'name', self.name)

    def fit(self, dmd: DMD, **kwargs):
        print('EstimatorDummy', 'fit', 'n_estimators', self.kwargs.get('n_estimators'))
        pass

    def predict(self, dmd: DMD, **kwargs):
        print('EstimatorDummy', 'predict', 'n_estimators', self.kwargs.get('n_estimators'))
        return DMD(x=dmd.values[:, 0:1])

    def predict_proba(self, dmd: DMD, **kwargs):
        print('EstimatorDummy', 'predict_proba', 'n_estimators', self.kwargs.get('n_estimators'))
        return DMD(x=dmd.values[:, 0:2])

class TestPyTrust(unittest.TestCase):

    def test_dummy(self):
        dmd_train = DMD(x=numpy.zeros((3, 3)), y=numpy.zeros((3, 1)))
        dmd_test = DMD(x=numpy.zeros((3, 3)), y=numpy.zeros((3, 1)))

        print('init model')
        model = DMDPipeline(transformers=[TransformerDummy],
                            estimator=EstimatorDummy,
                            hyperparameters=[{'name': 'Dummy'}, {'n_estimators': 10}])

        self.assertIsInstance(model.name, str)

        self.assertIsInstance(model.transformers, list)
        self.assertEqual(len(model.transformers), 1)
        self.assertIsInstance(model.transformers[0], TransformerBase)
        self.assertEqual(model.transformers[0].name, 'Dummy')

        self.assertIsInstance(model.estimator, EstimatorBase)
        self.assertEqual(model.estimator.kwargs['n_estimators'], 10)

        self.assertEqual(len(model.hyperparameters), 2)
        self.assertEqual((model.transformers[0].name, {'name': 'Dummy'}),
                         model.hyperparameters[0])
        self.assertEqual((model.estimator.name, {'n_estimators': 10}),
                         model.hyperparameters[1])


        print('fit model')
        model.fit(dmd_train)

        print('model predict')
        predictions = model.predict(dmd_test)
        self.assertIsInstance(predictions, DMD)



import os

import pandas
from sklearn.ensemble import RandomForestClassifier
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline

from pytolemaic import FeatureTypes, DMD, HOME_DIR, PyTrust


def run(fast=False):
    # read data
    train_path = os.path.join(HOME_DIR, 'resources', 'datasets', 'titanic-train.csv')

    df_train = pandas.read_csv(train_path)

    num, cat = FeatureTypes.numerical, FeatureTypes.categorical
    dmd_train, dmd_test = DMD.from_df(df_train=df_train, df_test=None,
                                      is_classification=True,
                                      target_name='Survived',
                                      feature_types=[num, cat, cat, cat, num, num, num, cat, num, cat, cat],
                                      categorical_encoding=True, nan_list=['?'],
                                      split_ratio=0.2)

    classifier = Pipeline(steps=[('Imputer', SimpleImputer()),
                                 ('Estimator', RandomForestClassifier(n_estimators=3))])

    classifier.fit(dmd_train.values, dmd_train.target)

    pytrust = PyTrust(
        model=classifier,
        xtrain=dmd_train,
        xtest=dmd_test,
        metric='recall')

    # some analysis
    print('\n'.join(pytrust.insights))

    pytrust.dataset_analysis_report.plot()
    pytrust.scoring_report.plot()
    pytrust.sensitivity_report.plot()


if __name__ == '__main__':
    from matplotlib import pyplot as plt

    run()
    plt.show()
